##
##

import attr
import attrs
from libcapella.logic.common import Audit, not_none


@attr.s
class AllowedCIDR:
    id: str = attr.ib()
    cidr: str = attr.ib()
    comment: str = attr.ib()
    expiresAt: str = attr.ib()
    status: str = attr.ib()
    type: str = attr.ib()
    audit: Audit = attr.ib()

    @classmethod
    def create(cls, data: dict):
        return cls(
            data.get("id"),
            data.get("cidr"),
            data.get("comment"),
            data.get("expiresAt"),
            data.get("status"),
            data.get("type"),
            Audit(
                data.get("audit", {}).get("createdBy"),
                data.get("audit", {}).get("createdAt"),
                data.get("audit", {}).get("modifiedBy"),
                data.get("audit", {}).get("modifiedAt"),
                data.get("audit", {}).get("version")
            )
        )

    @property
    def as_dict(self):
        # noinspection PyTypeChecker
        return attrs.asdict(self)

    @property
    def as_dict_striped(self):
        result = not_none(self.as_dict)
        if 'audit' in result:
            del result['audit']
        return result


class AllowedCIDRBuilder(object):

    def __init__(self,
                 cidr='0.0.0.0/0',
                 comment='Automation Created Allowed CIDR Block'):
        self._cidr = cidr
        self._comment = comment

    def cidr(self, cidr: str):
        self._cidr = cidr
        return self

    def comment(self, comment: str):
        self._comment = comment
        return self

    def build(self) -> AllowedCIDR:
        return AllowedCIDR.create(dict(
            cidr=self._cidr,
            comment=self._comment,
        ))



##
##

import logging
from typing import List, Union
from restfull.restapi import NotFoundError
from libcapella.organization import CapellaOrganization
from libcapella.logic.project import Project
from libcapella.user import CapellaUser

logger = logging.getLogger('libcapella.project')
logger.addHandler(logging.NullHandler())


class CapellaProject(object):

    def __init__(self, org: CapellaOrganization, project: str = None, email: str = None):
        self._endpoint = f"{org.endpoint}/{org.id}/projects"
        self.rest = org.rest
        self.project_name = project if project else org.project
        self.email = email
        self.org = org
        if self.project_name:
            self.project = self.get_by_name(self.project_name, self.email)
        else:
            self.project = None

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def id(self):
        if not self.project:
            return None
        return self.project.id

    def list(self) -> List[Project]:
        result = self.rest.get_paged(self._endpoint,
                                     total_tag="totalItems",
                                     pages_tag="last",
                                     per_page_tag="perPage",
                                     per_page=50,
                                     cursor="cursor",
                                     category="pages").validate().json_list()
        logger.debug(f"project list: found {result.size}")
        return [Project.create(r) for r in result.as_list]

    def owned_by_user(self, project_id: str, email: str) -> bool:
        user = CapellaUser(self.org, email)
        user_projects = user.projects_by_owner()
        return project_id in user_projects

    def get(self, project_id: str) -> Union[Project, None]:
        endpoint = self._endpoint + f"/{project_id}"
        try:
            result = self.rest.get(endpoint).validate().as_json().json_object()
            return Project.create(result.as_dict)
        except NotFoundError:
            return None

    def get_by_name(self, name: str, email: str = None) -> Union[Project, None]:
        result = self.rest.get_paged(self._endpoint,
                                     total_tag="totalItems",
                                     pages_tag="last",
                                     per_page_tag="perPage",
                                     per_page=50,
                                     cursor="cursor",
                                     category="pages").validate().filter("name", name).list_item(0)
        if not result:
            return None
        project = Project.create(result)
        if email and not self.owned_by_user(project.id, email):
            return None
        return project

    def create(self, project: Project):
        project_id = self.rest.post(self._endpoint, project.as_dict_striped).validate().as_json().json_key("id")
        project.id = project_id
        self.project = project



##
##

import attr
from libcapella.logic.common import Audit


@attr.s
class Preferences:
    sessionDuration: int = attr.ib()


@attr.s
class Organization:
    id: str = attr.ib()
    name: str = attr.ib()
    description: str = attr.ib()
    preferences: Preferences = attr.ib()
    audit: Audit = attr.ib()

    @classmethod
    def create(cls, data: dict):
        return cls(
            data.get("id"),
            data.get("name"),
            data.get("description"),
            Preferences(
                data.get("preferences").get("sessionDuration")
            ),
            Audit(
                data.get("audit").get("createdBy"),
                data.get("audit").get("createdAt"),
                data.get("audit").get("modifiedBy"),
                data.get("audit").get("modifiedAt"),
                data.get("audit").get("version")
            )
        )



##
##

import logging
from typing import List, Union
from restfull.restapi import NotFoundError
from libcapella.organization import CapellaOrganization
from libcapella.logic.user import User, ProjectOwnership

logger = logging.getLogger('libcapella.user')
logger.addHandler(logging.NullHandler())


class CapellaUser(object):

    def __init__(self, org: CapellaOrganization, email: Union[str, None] = None):
        self._endpoint = f"{org.endpoint}/{org.id}/users"
        self.rest = org.rest
        self.email = email
        if self.email:
            self.user_record = self.get_by_email(self.email)
        else:
            self.user_record = None

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def id(self):
        if not self.user_record:
            return None
        return self.user_record.id

    def list(self):
        return [User.create(r) for r in self.get_all_users()]

    def get_all_users(self) -> List[dict]:
        result = self.rest.get_paged(self._endpoint,
                                     total_tag="totalItems",
                                     pages_tag="last",
                                     per_page_tag="perPage",
                                     per_page=100,
                                     cursor="cursor",
                                     category="pages").validate().json_list()
        return result.as_list

    def get(self, user_id: str) -> Union[User, None]:
        endpoint = self._endpoint + f"/{user_id}"
        try:
            result = self.rest.get(endpoint).validate().as_json().json_object()
            return User.create(result.as_dict)
        except NotFoundError:
            return None

    def get_by_email(self, email: str) -> Union[User, None]:
        user = next((u for u in self.list() if u.email == email), None)
        while not user:
            user = next((u for u in self.list() if u.email == email), None)
        return user

    def set_project_owner(self, project_id: str):
        user_id = self.id
        endpoint = self._endpoint + f"/{user_id}"
        user_op = ProjectOwnership()
        user_op.add(project_id)
        self.rest.patch(endpoint, user_op.as_dict).validate()

    def projects_by_owner(self):
        if self.user_record:
            return [resource.id for resource in self.user_record.resources if resource.type == "project"]
        else:
            return []



##
##

import re
import logging
from typing import List, Union
from restfull.restapi import NotFoundError
from libcapella.logic.network_peers import NetworkPeers
from libcapella.database import CapellaDatabase

logger = logging.getLogger('libcapella.network_peers')
logger.addHandler(logging.NullHandler())


class CapellaNetworkPeers(object):

    def __init__(self, database: CapellaDatabase, name: str = "NetworkPeer"):
        self._endpoint = f"{database.endpoint}/{database.id}/networkPeers"
        self.rest = database.rest
        self.name = name
        self.network_peer = self.get_by_name(self.name)

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def id(self):
        if not self.network_peer:
            return None
        return self.network_peer.id

    @property
    def this(self) -> NetworkPeers:
        return self.network_peer

    @property
    def provider_id(self):
        return self.network_peer.providerConfig.providerId

    @property
    def hosted_zone_id(self):
        cmd_list = self.network_peer.commands if self.network_peer.commands else []
        return next((re.search('--hosted-zone-id=(.+?) ', z).group(1) for z in cmd_list if re.search('--hosted-zone-id', z)), None)

    @property
    def peer_project(self):
        cmd_list = self.network_peer.commands if self.network_peer.commands else []
        return next((re.search('--peer-project\s+(.+?) ', z).group(1) for z in cmd_list if re.search('--peer-project', z)), None)

    @property
    def peer_network(self):
        cmd_list = self.network_peer.commands if self.network_peer.commands else []
        return next((re.search('--peer-network\s+(.+?)$', z).group(1) for z in cmd_list if re.search('--peer-network', z)), None)

    @property
    def managed_zone(self):
        cmd_list = self.network_peer.commands if self.network_peer.commands else []
        return next((re.search('--dns-name=(.+?) ', z).group(1) for z in cmd_list if re.search('--dns-name', z)), None)

    def refresh(self):
        self.network_peer = self.get(self.network_peer.id)

    def list(self) -> List[NetworkPeers]:
        result = self.rest.get_paged(self._endpoint,
                                     total_tag="totalItems",
                                     pages_tag="last",
                                     per_page_tag="perPage",
                                     per_page=100,
                                     cursor="cursor",
                                     category="pages").validate().json_list()
        logger.debug(f"network peer list: found {result.size}")
        return [NetworkPeers.create(a) for a in result.as_list]

    def get(self, peer_id: str) -> Union[NetworkPeers, None]:
        endpoint = f"{self._endpoint}/{peer_id}"
        try:
            result = self.rest.get(endpoint).validate().as_json().json_object()
            return NetworkPeers.create(result.as_dict)
        except NotFoundError:
            return None

    def get_by_name(self, name: str) -> Union[NetworkPeers, None]:
        result = self.rest.get_paged(self._endpoint,
                                     total_tag="totalItems",
                                     pages_tag="last",
                                     per_page_tag="perPage",
                                     per_page=100,
                                     cursor="cursor",
                                     category="pages").validate().filter("name", name).list_item(0)
        if not result:
            return None
        return NetworkPeers.create(result)

    def create(self, network_peer: NetworkPeers):
        network_peer_id = self.rest.post(self._endpoint, network_peer.as_dict_striped).validate().as_json().json_key("id")
        network_peer.id = network_peer_id
        self.network_peer = network_peer

    def delete(self):
        if self.network_peer.id:
            endpoint = f"{self._endpoint}/{self.network_peer.id}"
            self.rest.delete(endpoint)



##
##

import logging
from typing import List
from libcapella.base import CouchbaseCapella
from libcapella.logic.organization import Organization

logger = logging.getLogger('libcapella.organization')
logger.addHandler(logging.NullHandler())


class CapellaOrganization(CouchbaseCapella):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._endpoint = "/v4/organizations"
        self.organization = self.get_default()

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def id_endpoint(self):
        return f"{self.endpoint}/{self.id}"

    @property
    def id(self):
        if not self.organization:
            return None
        return self.organization.id

    def list(self) -> List[Organization]:
        result = self.rest.get(self._endpoint).validate().as_json("data").json_list()
        logger.debug(f"organization list: found {result.size}")
        return [Organization.create(r) for r in result.as_list]

    def get(self, org_id: str) -> Organization:
        endpoint = self._endpoint + f"/{org_id}"
        result = self.rest.get(endpoint).validate().as_json().json_object()
        logger.debug(f"organization get:\n{result.formatted}")
        return Organization.create(result.as_dict)

    def get_default(self) -> Organization:
        result = self.list()
        if not len(result) >= 1:
            raise RuntimeError("No organizations found")
        return result[0]



##
##

import attr
import attrs
from typing import List
from libcapella.logic.common import Audit, not_none


@attr.s
class AzureConfig:
    azureTenantId = attr.ib()
    subscriptionId = attr.ib()
    resourceGroup = attr.ib()
    vnetId = attr.ib()
    cidr = attr.ib()

    @property
    def as_dict(self):
        # noinspection PyTypeChecker
        return attrs.asdict(self)


@attr.s
class GCPConfig:
    networkName = attr.ib()
    cidr = attr.ib()
    projectId = attr.ib()
    serviceAccount = attr.ib()

    @property
    def as_dict(self):
        # noinspection PyTypeChecker
        return attrs.asdict(self)


@attr.s
class AWSConfig:
    accountId = attr.ib()
    vpcId = attr.ib()
    region = attr.ib()
    cidr = attr.ib()

    @property
    def as_dict(self):
        # noinspection PyTypeChecker
        return attrs.asdict(self)


@attr.s
class ProviderConfig:
    providerId = attr.ib()
    AWSConfig = attr.ib(default=None)
    GCPConfig = attr.ib(default=None)
    AzureConfig = attr.ib(default=None)


@attr.s
class NetworkPeerStatus:
    state = attr.ib()
    reasoning = attr.ib()


@attr.s
class NetworkPeers:
    id: str = attr.ib()
    name: str = attr.ib()
    providerType: str = attr.ib()
    status: NetworkPeerStatus = attr.ib()
    commands: List[str] = attr.ib()
    providerConfig: ProviderConfig = attr.ib()
    audit: Audit = attr.ib()

    @classmethod
    def create(cls, data: dict, new: bool = False):
        return cls(
            data.get("id"),
            data.get("name"),
            data.get("providerType"),
            NetworkPeerStatus(
                data.get("status", {}).get("state"),
                data.get("status", {}).get("reasoning"),
            ),
            data.get("commands", []),
            ProviderConfig(
                data.get("providerConfig", {}).get("providerId"),
                AWSConfig(
                    data.get("providerConfig").get("AWSConfig").get("accountId"),
                    data.get("providerConfig").get("AWSConfig").get("vpcId"),
                    data.get("providerConfig").get("AWSConfig").get("region"),
                    data.get("providerConfig").get("AWSConfig").get("cidr"),
                ) if data.get("providerConfig", {}).get("AWSConfig") else None,
                GCPConfig(
                    data.get("providerConfig").get("GCPConfig").get("networkName"),
                    data.get("providerConfig").get("GCPConfig").get("cidr"),
                    data.get("providerConfig").get("GCPConfig").get("projectId"),
                    data.get("providerConfig").get("GCPConfig").get("serviceAccountId"),
                ) if data.get("providerConfig", {}).get("GCPConfig") else None,
                AzureConfig(
                    data.get("providerConfig").get("AzureConfig").get("azureTenantId"),
                    data.get("providerConfig").get("AzureConfig").get("subscriptionId"),
                    data.get("providerConfig").get("AzureConfig").get("resourceGroup"),
                    data.get("providerConfig").get("AzureConfig").get("vnetId"),
                    data.get("providerConfig").get("AzureConfig").get("cidr"),
                ) if data.get("providerConfig", {}).get("AzureConfig") else None,
            ) if not new else data.get("providerConfig", {}),
            Audit(
                data.get("audit", {}).get("createdBy"),
                data.get("audit", {}).get("createdAt"),
                data.get("audit", {}).get("modifiedBy"),
                data.get("audit", {}).get("modifiedAt"),
                data.get("audit", {}).get("version")
            )
        )

    @property
    def as_dict(self):
        # noinspection PyTypeChecker
        return attrs.asdict(self)

    @property
    def as_dict_striped(self):
        result = not_none(self.as_dict)
        if 'audit' in result:
            del result['audit']
        if 'status' in result and not result['status']:
            del result['status']
        if 'commands' in result and len(result['commands']) == 0:
            del result['commands']
        return result


class NetworkPeerBuilder(object):

    def __init__(self):
        self._provider_type = "aws"
        self._name = "NetworkPeer"
        self._account_id = None
        self._vpc_id = None
        self._region = None
        self._cidr = None
        self._network_name = None
        self._project_id = None
        self._service_account = None
        self._tenant_id = None
        self._subscription_id = None
        self._resource_group = None
        self._vnet_id = None

    def provider_type(self, provider_type: str):
        self._provider_type = provider_type
        return self

    def account_id(self, account_id: str):
        self._account_id = account_id
        return self

    def vpc_id(self, vpc_id: str):
        self._vpc_id = vpc_id
        return self

    def region(self, region: str):
        self._region = region
        return self

    def cidr(self, cidr: str):
        self._cidr = cidr
        return self

    def network_name(self, network_name: str):
        self._network_name = network_name
        return self

    def project_id(self, project_id: str):
        self._project_id = project_id
        return self

    def service_account(self, service_account: str):
        self._service_account = service_account
        return self

    def tenant_id(self, tenant_id: str):
        self._tenant_id = tenant_id
        return self

    def subscription_id(self, subscription_id: str):
        self._subscription_id = subscription_id
        return self

    def resource_group(self, resource_group: str):
        self._resource_group = resource_group
        return self

    def vnet_id(self, vnet_id: str):
        self._vnet_id = vnet_id
        return self

    def build(self) -> NetworkPeers:
        provider_config = {}
        if self._provider_type == "aws":
            provider_config = AWSConfig(
                self._account_id,
                self._vpc_id,
                self._region,
                self._cidr,
            ).as_dict
        elif self._provider_type == "gcp":
            provider_config = GCPConfig(
                self._network_name,
                self._cidr,
                self._project_id,
                self._service_account,
            ).as_dict
        elif self._provider_type == "azure":
            provider_config = AzureConfig(
                self._tenant_id,
                self._subscription_id,
                self._resource_group,
                self._vnet_id,
                self._cidr,
            )
        return NetworkPeers.create(dict(
            name=self._name,
            providerType=self._provider_type,
            providerConfig=provider_config
        ), new=True)



##
##

import attr


def not_none(data):
    if isinstance(data, dict):
        for key, value in list(data.items()):
            if isinstance(value, (list, dict, tuple, set)):
                data[key] = not_none(value)
            elif value is None or key is None:
                del data[key]

    elif isinstance(data, (list, set, tuple)):
        data = type(data)(not_none(item) for item in data if item is not None)

    return data


@attr.s
class Audit:
    createdBy: str = attr.ib()
    createdAt: str = attr.ib()
    modifiedBy: str = attr.ib()
    modifiedAt: str = attr.ib()
    version: int = attr.ib()



##
##

import time
import logging
from typing import List, Union
from restfull.restapi import NotFoundError
from libcapella.project import CapellaProject
from libcapella.logic.columnar import Columnar

logger = logging.getLogger('libcapella.columnar')
logger.addHandler(logging.NullHandler())


class CapellaColumnar(object):

    def __init__(self, project: CapellaProject, cluster: Union[str, None] = None):
        self._endpoint = f"{project.endpoint}/{project.id}/analyticsClusters"
        self.rest = project.rest
        self.project = project
        self.cluster_name = cluster
        if self.cluster_name:
            self.cluster = self.get_by_name(self.cluster_name)
        else:
            self.cluster = None

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def id(self):
        if not self.cluster:
            return None
        return self.cluster.id

    @property
    def this(self) -> Columnar:
        return self.cluster

    def refresh(self):
        self.cluster = self.get(self.cluster.id)

    def list(self) -> List[Columnar]:
        result = self.rest.get_paged(self._endpoint,
                                     total_tag="totalItems",
                                     pages_tag="last",
                                     per_page_tag="perPage",
                                     per_page=50,
                                     cursor="cursor",
                                     category="pages").validate().json_list()
        logger.debug(f"database list: found {result.size}")
        return [Columnar.create(r) for r in result.as_list]

    def get(self, columnar_id: str) -> Union[Columnar, None]:
        if not columnar_id:
            return None
        endpoint = f"{self._endpoint}/{columnar_id}"
        try:
            result = self.rest.get(endpoint).validate().as_json().json_object()
            return Columnar.create(result.as_dict)
        except NotFoundError:
            return None

    def get_by_name(self, name: str) -> Union[Columnar, None]:
        result = self.rest.get_paged(self._endpoint,
                                     total_tag="totalItems",
                                     pages_tag="last",
                                     per_page_tag="perPage",
                                     per_page=50,
                                     cursor="cursor",
                                     category="pages").validate().filter("name", name).list_item(0)
        if not result:
            return None
        return Columnar.create(result)

    def create(self, columnar: Columnar):
        cluster_id = self.rest.post(self._endpoint, columnar.as_dict_striped).validate().as_json().json_key("id")
        columnar.id = cluster_id
        self.cluster = columnar

    def wait(self, state, retry_count: int = 90):
        if not self.cluster:
            return True
        for retry_number in range(retry_count + 1):
            check = self.get(self.cluster.id)
            logger.debug(f"Checking cluster state {check.currentState if check else None} with state {state}")
            if check and check.currentState != state:
                return True
            elif not check:
                return True
            else:
                if retry_number == retry_count:
                    return False
                logger.debug(f"Waiting for cluster {self.cluster.name} to reach state {state}")
                time.sleep(10)

    def delete(self):
        if self.cluster.id:
            endpoint = f"{self._endpoint}/{self.cluster.id}"
            self.rest.delete(endpoint)



##
##

import attr
import attrs
from enum import Enum
from typing import List, Union
from libcapella.logic.common import Audit, not_none

aws_storage_matrix = {
    99: 3000,
    199: 4370,
    299: 5740,
    399: 7110,
    499: 8480,
    599: 9850,
    699: 11220,
    799: 12590,
    899: 13960,
    999: 15330,
    16384: 16000
}

azure_storage_matrix = {
    64: "P6",
    128: "P10",
    256: "P15",
    512: "P20",
    1024: "P30",
    2048: "P40",
    4096: "P50",
    8192: "P60"
}

azure_ultra_matrix = {
    64: 3000,
    128: 4000,
    256: 6000,
    512: 8000,
    1024: 16000,
    2048: 16000,
    3072: 16000,
    4096: 16000,
    5120: 16000,
    6144: 16000,
    7168: 16000,
    8192: 16000,
    9216: 16000,
    10240: 16000,
    11264: 16000,
    12288: 16000,
    13312: 16000,
    14336: 16000,
    15360: 16000
}


@attr.s
class CloudProvider:
    type: str = attr.ib()
    region: str = attr.ib()
    cidr: str = attr.ib()


@attr.s
class CouchbaseServer:
    version: str = attr.ib()


@attr.s
class ComputeConfig:
    cpu: int = attr.ib()
    ram: int = attr.ib()


@attr.s
class StorageConfig:
    storage: int = attr.ib()
    type: str = attr.ib()
    iops: int = attr.ib()


@attr.s
class NodeConfig:
    compute: ComputeConfig = attr.ib()
    disk: StorageConfig = attr.ib()


@attr.s
class ServiceGroup:
    node: NodeConfig = attr.ib()
    numOfNodes: int = attr.ib()
    services: List[str] = attr.ib()


class NodeAvailability(str, Enum):
    single = 'single'
    multi = 'multi'


class SupportPlan(str, Enum):
    basic = 'basic'
    devpro = 'developer pro'
    enterprise = 'enterprise'


class SupportTZ(str, Enum):
    eastern_us = 'ET'
    emea = 'GMT'
    asia = 'IST'
    western_us = 'PT'


@attr.s
class Availability:
    type: NodeAvailability = attr.ib()


@attr.s
class Support:
    plan: SupportPlan = attr.ib()
    timezone: SupportTZ = attr.ib()


@attr.s
class Database:
    id: str = attr.ib()
    appServiceId: str = attr.ib()
    name: str = attr.ib()
    description: str = attr.ib()
    configurationType: str = attr.ib()
    connectionString: str = attr.ib()
    cloudProvider: CloudProvider = attr.ib()
    couchbaseServer: CouchbaseServer = attr.ib()
    serviceGroups: List[ServiceGroup] = attr.ib()
    availability: Availability = attr.ib()
    support: Support = attr.ib()
    currentState: str = attr.ib()
    audit: Audit = attr.ib()
    cmekId: str = attr.ib()

    @classmethod
    def create(cls, data: dict):
        return cls(
            data.get("id"),
            data.get("appServiceId"),
            data.get("name"),
            data.get("description"),
            data.get("configurationType"),
            data.get("connectionString"),
            CloudProvider(
                data.get("cloudProvider", {}).get("type"),
                data.get("cloudProvider", {}).get("region"),
                data.get("cloudProvider", {}).get("cidr"),
            ),
            CouchbaseServer(
                data.get("couchbaseServer", {}).get("version"),
            ),
            [
                ServiceGroup(
                    NodeConfig(
                        ComputeConfig(
                            g.get("node", {}).get("compute", {}).get("cpu"),
                            g.get("node", {}).get("compute", {}).get("ram"),
                        ),
                        StorageConfig(
                            g.get("node", {}).get("disk", {}).get("storage"),
                            g.get("node", {}).get("disk", {}).get("type"),
                            g.get("node", {}).get("disk", {}).get("iops"),
                        )
                    ),
                    g.get("numOfNodes"),
                    g.get("services"),
                ) for g in data.get("serviceGroups", [])
            ],
            Availability(
                NodeAvailability(data.get("availability", {}).get("type", "multi"))
            ),
            Support(
                SupportPlan(data.get("support", {}).get("plan", "developer pro")),
                SupportTZ(data.get("support", {}).get("timezone", "PT"))
            ),
            data.get("currentState"),
            Audit(
                data.get("audit", {}).get("createdBy"),
                data.get("audit", {}).get("createdAt"),
                data.get("audit", {}).get("modifiedBy"),
                data.get("audit", {}).get("modifiedAt"),
                data.get("audit", {}).get("version")
            ),
            data.get("cmekId"),
        )

    @property
    def as_dict(self):
        # noinspection PyTypeChecker
        return attrs.asdict(self)

    @property
    def as_dict_striped(self):
        result = not_none(self.as_dict)
        if 'audit' in result:
            del result['audit']
        return result


class CapellaDatabaseBuilder(object):

    def __init__(self,
                 cloud,
                 name='cbdb',
                 description='Couchbase Cluster',
                 region='us-east-1',
                 cidr='10.1.20.0/23',
                 availability='multi',
                 plan='developer',
                 timezone='us_west',
                 version=None):
        self._name = name
        self._description = description
        self._cloud = cloud
        self._region = region
        self._service_groups = []
        self._cidr = cidr
        self._availability = availability
        self._plan = plan
        self._timezone = timezone
        self._version = version

    def name(self, name: str):
        self._name = name
        return self

    def description(self, description: str):
        self._description = description
        return self

    def cloud(self, cloud: str):
        self._cloud = cloud
        return self

    def region(self, region: str):
        self._region = region
        return self

    def cidr(self, cidr: str):
        self._cidr = cidr
        return self

    def availability(self, availability: str):
        self._availability = availability
        return self

    def plan(self, plan: str):
        self._plan = plan
        return self

    def timezone(self, timezone: str):
        self._timezone = timezone
        return self

    def version(self, version: str):
        self._version = version
        return self

    def service_group(self, machine_type: str, quantity: Union[str, int], storage: Union[str, int], services: Union[None, str, List[str]] = None, ultra: bool = False):
        cpu, ram = machine_type.split('x')
        _cpu = int(cpu)
        _ram = int(ram)

        if self._cloud == "aws":
            _storage = int(storage) if quantity > 1 else 100
            _iops = next((aws_storage_matrix[s] for s in aws_storage_matrix if s >= _storage), None)
            _storage_type = "gp3"
        elif self._cloud == "azure":
            if not ultra:
                if quantity > 1:
                    _, s_type = next(((s, azure_storage_matrix[s]) for s in azure_storage_matrix if s >= int(storage)), None)
                else:
                    _, s_type = 128, "P10"
                _storage = None
                _storage_type = s_type
                _iops = None
            else:
                if quantity == 1:
                    raise ValueError("Ultra disks are not compatible with single node clusters")
                size, iops = next(((s, azure_ultra_matrix[s]) for s in azure_ultra_matrix if s >= int(storage)), None)
                _storage = size
                _storage_type = "Ultra"
                _iops = iops
        else:
            _storage = int(storage) if quantity > 1 else 100
            _storage_type = "pd-ssd"
            _iops = None

        _quantity = int(quantity)

        if not services:
            _services = ["data", "query", "index", "search"]
        elif type(services) is str and services == "default":
            _services = ["data", "query", "index", "search"]
        else:
            _services = services

        service_group = dict(
            cpu=_cpu,
            ram=_ram,
            storage=_storage,
            iops=_iops,
            storage_type=_storage_type,
            quantity=_quantity,
            services=_services,
        )
        self._service_groups.append(service_group)

        return self

    def build(self) -> Database:
        quantity = next((g.get('quantity') for g in self._service_groups if g.get('quantity') == 1), 3)
        if quantity > 1:
            availability = "multi"
        else:
            availability = "single"

        if not self._plan or self._plan == "developer" or self._plan == "devpro" or self._plan == "developer-pro" or self._plan == "developer pro":
            plan = "developer pro"
        elif self._plan == "enterprise" or self._plan == "prod" or self._plan == "production":
            plan = "enterprise"
        else:
            plan = "basic"

        if not self._timezone or self._timezone == "us_west" or self._timezone == "west":
            timezone = "PT"
        elif self._timezone == "us_east" or self._timezone == "east":
            timezone = "ET"
        elif self._timezone == "europe" or self._timezone == "UTC" or self._timezone.lower() == "emea":
            timezone = "GMT"
        else:
            timezone = "IST"

        return Database.create(dict(
            name=self._name,
            description=self._description,
            cloudProvider=dict(
                type=self._cloud,
                region=self._region,
                cidr=self._cidr,
            ),
            couchbaseServer=dict(
                version=self._version
            ),
            serviceGroups=[
                dict(
                    node=dict(
                        compute=dict(
                            cpu=g.get("cpu"),
                            ram=g.get("ram")
                        ),
                        disk=dict(
                            storage=g.get("storage"),
                            type=g.get("storage_type"),
                            iops=g.get("iops")
                        )
                    ),
                    numOfNodes=g.get("quantity"),
                    services=g.get("services")
                ) for g in self._service_groups
            ],
            availability=dict(
                type=availability,
            ),
            support=dict(
                plan=plan,
                timezone=timezone,
            )
        ))



##
##

import logging
from libcapella.config import CapellaConfig
from restfull.restapi import RestAPI
from restfull.bearer_auth import BearerAuth

logger = logging.getLogger('libcapella.base')
logger.addHandler(logging.NullHandler())


class CouchbaseCapella(object):

    def __init__(self, config: CapellaConfig):
        self.auth_token = config.token
        self.api_host = config.api_host
        self.project = config.project
        logger.debug(f"using auth profile {config.profile} API key ID {config.key_id}")

        auth = BearerAuth(self.auth_token)
        self.rest = RestAPI(auth, self.api_host)



##
##

import logging
from typing import List, Union
from restfull.restapi import NotFoundError
from libcapella.logic.allowed_cidr import AllowedCIDR
from libcapella.columnar import CapellaColumnar

logger = logging.getLogger('libcapella.columnar_allowed_cidr')
logger.addHandler(logging.NullHandler())


class ColumnarAllowedCIDR(object):

    def __init__(self, columnar: CapellaColumnar, cidr: str = None):
        self._endpoint = f"{columnar.endpoint}/{columnar.id}/allowedcidrs"
        self.rest = columnar.rest
        self.cidr = cidr
        if self.cidr:
            self.allowed_cidr = self.get_by_name(self.cidr)
        else:
            self.allowed_cidr = None

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def id(self):
        if not self.allowed_cidr:
            return None
        return self.allowed_cidr.id

    def list(self) -> List[AllowedCIDR]:
        result = self.rest.get_paged(self._endpoint,
                                     total_tag="totalItems",
                                     pages_tag="last",
                                     per_page_tag="perPage",
                                     per_page=100,
                                     cursor="cursor",
                                     category="pages").validate().json_list()
        logger.debug(f"allowed CIDR list: found {result.size}")
        return [AllowedCIDR.create(a) for a in result.as_list]

    def get(self, allowed_cidr_id: str) -> Union[AllowedCIDR, None]:
        endpoint = f"{self._endpoint}/{allowed_cidr_id}"
        try:
            result = self.rest.get(endpoint).validate().as_json().json_object()
            return AllowedCIDR.create(result.as_dict)
        except NotFoundError:
            return None

    def get_by_name(self, cidr: str) -> Union[AllowedCIDR, None]:
        result = self.rest.get_paged(self._endpoint,
                                     total_tag="totalItems",
                                     pages_tag="last",
                                     per_page_tag="perPage",
                                     per_page=100,
                                     cursor="cursor",
                                     category="pages").validate().filter("cidr", cidr).list_item(0)
        if not result:
            return None
        return AllowedCIDR.create(result)

    def create(self, allowed_cidr: AllowedCIDR):
        allowed_cidr_id = self.rest.post(self._endpoint, allowed_cidr.as_dict_striped).validate().as_json().json_key("id")
        allowed_cidr.id = allowed_cidr_id
        self.allowed_cidr = allowed_cidr

    def delete(self):
        if self.allowed_cidr.id:
            endpoint = f"{self._endpoint}/{self.allowed_cidr.id}"
            self.rest.delete(endpoint)



##
##

import attr
import attrs
from enum import Enum
from typing import Union
from libcapella.logic.common import not_none


@attr.s
class ComputeConfig:
    cpu: int = attr.ib()
    ram: int = attr.ib()


class NodeAvailability(str, Enum):
    single = 'single'
    multi = 'multi'


class SupportPlan(str, Enum):
    devpro = 'developer pro'
    enterprise = 'enterprise'


class SupportTZ(str, Enum):
    eastern_us = 'ET'
    emea = 'GMT'
    asia = 'IST'
    western_us = 'PT'


@attr.s
class Availability:
    type: NodeAvailability = attr.ib()


@attr.s
class Support:
    plan: SupportPlan = attr.ib()
    timezone: SupportTZ = attr.ib()


@attr.s
class Columnar:
    id: str = attr.ib()
    name: str = attr.ib()
    description: str = attr.ib()
    cloudProvider: str = attr.ib()
    region: str = attr.ib()
    nodes: int = attr.ib()
    currentState: str = attr.ib()
    support: Support = attr.ib()
    compute: ComputeConfig = attr.ib()
    availability: Availability = attr.ib()

    @classmethod
    def create(cls, data: dict):
        return cls(
            data.get("id"),
            data.get("name"),
            data.get("description"),
            data.get("cloudProvider"),
            data.get("region"),
            data.get("nodes"),
            data.get("currentState"),
            Support(
                SupportPlan(data.get("support", {}).get("plan", "developer pro")),
                SupportTZ(data.get("support", {}).get("timezone", "PT"))
            ),
            ComputeConfig(
                data.get("compute", {}).get("cpu"),
                data.get("compute", {}).get("ram"),
            ),
            Availability(
                NodeAvailability(data.get("availability", {}).get("type", "multi"))
            )
        )

    @property
    def as_dict(self):
        # noinspection PyTypeChecker
        return attrs.asdict(self)

    @property
    def as_dict_striped(self):
        result = not_none(self.as_dict)
        if 'audit' in result:
            del result['audit']
        return result


class CapellaColumnarBuilder(object):

    def __init__(self,
                 cloud,
                 name='datamart',
                 description='Columnar Cluster',
                 cpu=4,
                 ram=32,
                 quantity=1,
                 region='us-east-1',
                 availability='single',
                 plan='developer',
                 timezone='us_west'):
        self._name = name
        self._description = description
        self._cpu = cpu
        self._ram = ram
        self._quantity = quantity
        self._cloud = cloud
        self._region = region
        self._availability = availability
        self._plan = plan
        self._timezone = timezone

    def name(self, name: str):
        self._name = name
        return self

    def description(self, description: str):
        self._description = description
        return self

    def cloud(self, cloud: str):
        self._cloud = cloud
        return self

    def region(self, region: str):
        self._region = region
        return self

    def availability(self, availability: str):
        self._availability = availability
        return self

    def plan(self, plan: str):
        self._plan = plan
        return self

    def timezone(self, timezone: str):
        self._timezone = timezone
        return self

    def compute(self, machine_type: str, quantity: Union[str, int]):
        cpu, ram = machine_type.split('x')
        _cpu = int(cpu)
        _ram = int(ram)
        _quantity = int(quantity)
        self._cpu = _cpu
        self._ram = _ram
        self._quantity = _quantity
        return self

    def build(self) -> Columnar:
        if not self._availability or self._availability == "single":
            availability = "single"
        else:
            availability = "multi"

        if not self._plan or self._plan == "developer" or self._plan == "devpro" or self._plan == "developer-pro" or self._plan == "developer pro":
            plan = "developer pro"
        else:
            plan = "enterprise"

        if not self._timezone or self._timezone == "us_west" or self._timezone == "west":
            timezone = "PT"
        elif self._timezone == "us_east" or self._timezone == "east":
            timezone = "ET"
        elif self._timezone == "europe" or self._timezone == "UTC" or self._timezone.lower() == "emea":
            timezone = "GMT"
        else:
            timezone = "IST"

        return Columnar.create(dict(
            name=self._name,
            description=self._description,
            cloudProvider=self._cloud,
            region=self._region,
            nodes=self._quantity,
            support=dict(
                plan=plan,
                timezone=timezone,
            ),
            compute=dict(
                cpu=self._cpu,
                ram=self._ram,
            ),
            availability=dict(
                type=availability,
            )
        ))



##
##

import attr
import attrs
from typing import List
from libcapella.logic.common import Audit, not_none


@attr.s
class ResourceScope:
    name: str = attr.ib()
    collections: List[str] = attr.ib()


@attr.s
class BucketAccessSpec:
    name: str = attr.ib()
    scopes: List[ResourceScope] = attr.ib()


@attr.s
class ResourceBucket:
    buckets: List[BucketAccessSpec] = attr.ib()


@attr.s
class Access:
    privileges: List[str] = attr.ib()
    resources: ResourceBucket = attr.ib(default={})


@attr.s
class DatabaseCredentials:
    id: str = attr.ib()
    name: str = attr.ib()
    password: str = attr.ib()
    audit: Audit = attr.ib()
    access: List[Access] = attr.ib()

    @classmethod
    def create(cls, data: dict):
        return cls(
            data.get("id"),
            data.get("name"),
            data.get("password"),
            Audit(
                data.get("audit", {}).get("createdBy"),
                data.get("audit", {}).get("createdAt"),
                data.get("audit", {}).get("modifiedBy"),
                data.get("audit", {}).get("modifiedAt"),
                data.get("audit", {}).get("version")
            ),
            [
                Access(
                    a.get("privileges", []),
                    ResourceBucket(
                        [
                            BucketAccessSpec(
                                b.get("name"),
                                [
                                    ResourceScope(
                                        s.get("name"),
                                        s.get("collections", []),
                                    ) for s in b.get("scopes", [])
                                ]
                            ) for b in a.get("resources", {}).get("buckets", [])
                        ]
                    )
                ) for a in data.get("access", [])
            ]
        )

    @property
    def as_dict(self):
        # noinspection PyTypeChecker
        return attrs.asdict(self)

    @property
    def as_dict_striped(self):
        result = not_none(self.as_dict)
        if 'audit' in result:
            del result['audit']
        if len(result.get('access')[0].get('resources', {}).get('buckets', [])) == 0:
            del result['access'][0]['resources']
        return result


class DatabaseCredentialsBuilder(object):

    def __init__(self,
                 username='developer',
                 password='P@ssw0rd!',
                 ):
        self._username = username
        self._password = password
        self._privileges = []

    def username(self, username: str):
        self._username = username
        return self

    def password(self, password: str):
        self._password = password
        return self

    def data_reader(self):
        self._privileges.append("data_reader")

    def data_writer(self):
        self._privileges.append("data_writer")

    def data_read_write(self):
        self._privileges.extend(["data_reader", "data_writer"])

    def build(self) -> DatabaseCredentials:
        return DatabaseCredentials.create(dict(
            name=self._username,
            password=self._password,
            access=[
                dict(
                    privileges=self._privileges,
                )
            ]
        ))



##
##

import attr
import attrs
from typing import Union
from libcapella.logic.common import Audit, not_none


@attr.s
class ComputeConfig:
    cpu: int = attr.ib()
    ram: int = attr.ib()


@attr.s
class AppService:
    id: str = attr.ib()
    name: str = attr.ib()
    description: str = attr.ib()
    cloudProvider: str = attr.ib()
    nodes: int = attr.ib()
    compute: ComputeConfig = attr.ib()
    clusterId: str = attr.ib()
    currentState: str = attr.ib()
    version: str = attr.ib()
    audit: Audit = attr.ib()

    @classmethod
    def create(cls, data: dict):
        return cls(
            data.get("id"),
            data.get("name"),
            data.get("description"),
            data.get("cloudProvider"),
            data.get("nodes"),
            ComputeConfig(
                data.get("compute", {}).get("cpu"),
                data.get("compute", {}).get("ram"),
            ),
            data.get("clusterId"),
            data.get("currentState"),
            data.get("version"),
            Audit(
                data.get("audit", {}).get("createdBy"),
                data.get("audit", {}).get("createdAt"),
                data.get("audit", {}).get("modifiedBy"),
                data.get("audit", {}).get("modifiedAt"),
                data.get("audit", {}).get("version")
            )
        )

    @property
    def as_dict(self):
        # noinspection PyTypeChecker
        return attrs.asdict(self)

    @property
    def as_dict_striped(self):
        result = not_none(self.as_dict)
        if 'audit' in result:
            del result['audit']
        return result


class CapellaAppServiceBuilder(object):

    def __init__(self):
        self._name = "AppService"
        self._description = "Automation Generated App Service"
        self._version = None
        self._nodes = 2
        self._cpu = 8
        self._ram = 16

    def name(self, name: str):
        self._name = name
        return self

    def description(self, description: str):
        self._description = description
        return self

    def version(self, version: str):
        self._version = version
        return self

    def compute(self, machine_type: str, nodes: Union[str, int]):
        cpu, ram = machine_type.split('x')
        _cpu = int(cpu)
        _ram = int(ram)
        _nodes = int(nodes)
        self._cpu = _cpu
        self._ram = _ram
        self._nodes = _nodes
        return self

    def build(self) -> AppService:
        return AppService.create(dict(
            name=self._name,
            description=self._description,
            nodes=self._nodes,
            compute=dict(
                cpu=self._cpu,
                ram=self._ram,
            ),
            version=self._version
        ))



##
##

import time
import logging
from typing import List, Union
from restfull.restapi import NotFoundError
from libcapella.logic.app_service import AppService
from libcapella.database import CapellaDatabase

logger = logging.getLogger('libcapella.app_service')
logger.addHandler(logging.NullHandler())


class CapellaAppService(object):

    def __init__(self, database: CapellaDatabase):
        self._endpoint = f"{database.endpoint}/{database.id}/appservices"
        self._list_endpoint = f"{database.project.org.id_endpoint}/appservices"
        self.database = database
        self.rest = database.rest
        self.app_service = self.get_by_db(database.id)

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def id(self):
        if not self.app_service:
            return None
        return self.app_service.id

    def list(self) -> List[AppService]:
        result = self.rest.get_paged(self._list_endpoint,
                                     total_tag="totalItems",
                                     pages_tag="last",
                                     per_page_tag="perPage",
                                     per_page=100,
                                     cursor="cursor",
                                     category="pages").validate().json_list()
        logger.debug(f"app service list: found {result.size}")
        return [AppService.create(a) for a in result.as_list]

    def get(self, app_service_id: str) -> Union[AppService, None]:
        if not app_service_id:
            return None
        endpoint = f"{self._endpoint}/{app_service_id}"
        try:
            result = self.rest.get(endpoint).validate().as_json().json_object()
            return AppService.create(result.as_dict)
        except NotFoundError:
            return None

    def get_by_db(self, database_id: str) -> Union[AppService, None]:
        if not database_id:
            return None
        result = self.rest.get_paged(self._list_endpoint,
                                     total_tag="totalItems",
                                     pages_tag="last",
                                     per_page_tag="perPage",
                                     per_page=100,
                                     cursor="cursor",
                                     category="pages").validate().filter("clusterId", database_id).list_item(0)
        if not result:
            return None
        return AppService.create(result)

    def create(self, app_service: AppService):
        app_service_id = self.rest.post(self._endpoint, app_service.as_dict_striped).validate().as_json().json_key("id")
        app_service.id = app_service_id
        self.app_service = app_service

    def wait(self, state, until: bool = False, retry_count: int = 90):
        if not self.app_service:
            return True
        for retry_number in range(retry_count + 1):
            check = self.get(self.app_service.id)
            logger.debug(f"Checking app service state {check.currentState if check else None} with state {state} until {until}")
            if not until and check and check.currentState != state:
                return True
            elif until and check and check.currentState == state:
                return True
            elif not check:
                return True
            else:
                if retry_number == retry_count:
                    return False
                logger.debug(f"Waiting for app service {self.app_service.name} to reach state {state}")
                time.sleep(10)

    def delete(self):
        if self.app_service.id:
            endpoint = f"{self._endpoint}/{self.app_service.id}"
            self.rest.delete(endpoint)



##
##

import logging
from typing import List, Union
from restfull.restapi import NotFoundError
from libcapella.logic.credentials import DatabaseCredentials
from libcapella.database import CapellaDatabase

logger = logging.getLogger('libcapella.database_credentials')
logger.addHandler(logging.NullHandler())


class CapellaDatabaseCredentials(object):

    def __init__(self, database: CapellaDatabase, username: str = None):
        self._endpoint = f"{database.endpoint}/{database.id}/users"
        self.rest = database.rest
        self.username = username
        if self.username:
            self.db_credentials = self.get_by_name(self.username)
        else:
            self.db_credentials = None

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def id(self):
        if not self.db_credentials:
            return None
        return self.db_credentials.id

    def list(self) -> List[DatabaseCredentials]:
        result = self.rest.get_paged(self._endpoint,
                                     total_tag="totalItems",
                                     pages_tag="last",
                                     per_page_tag="perPage",
                                     per_page=100,
                                     cursor="cursor",
                                     category="pages").validate().json_list()
        logger.debug(f"database credentials list: found {result.size}")
        return [DatabaseCredentials.create(a) for a in result.as_list]

    def get(self, db_user_id: str) -> Union[DatabaseCredentials, None]:
        endpoint = f"{self._endpoint}/{db_user_id}"
        try:
            result = self.rest.get(endpoint).validate().as_json().json_object()
            return DatabaseCredentials.create(result.as_dict)
        except NotFoundError:
            return None

    def get_by_name(self, username: str) -> Union[DatabaseCredentials, None]:
        result = self.rest.get_paged(self._endpoint,
                                     total_tag="totalItems",
                                     pages_tag="last",
                                     per_page_tag="perPage",
                                     per_page=100,
                                     cursor="cursor",
                                     category="pages").validate().filter("name", username).list_item(0)
        if not result:
            return None
        return DatabaseCredentials.create(result)

    def create(self, db_user: DatabaseCredentials):
        db_user_id = self.rest.post(self._endpoint, db_user.as_dict_striped).validate().as_json().json_key("id")
        db_user.id = db_user_id
        self.db_credentials = db_user

    def delete(self):
        if self.db_credentials.id:
            endpoint = f"{self._endpoint}/{self.db_credentials.id}"
            self.rest.delete(endpoint)



##
##

import attr
import attrs
from typing import Optional, List
from libcapella.logic.common import Audit


@attr.s
class Resources:
    type: str = attr.ib()
    id: str = attr.ib()
    roles: List[str] = attr.ib()


@attr.s
class User:
    id: str = attr.ib()
    name: str = attr.ib()
    email: str = attr.ib()
    status: str = attr.ib()
    inactive: bool = attr.ib()
    organizationId: str = attr.ib()
    organizationRoles: List[str] = attr.ib()
    lastLogin: str = attr.ib()
    region: str = attr.ib()
    timeZone: str = attr.ib()
    enableNotifications: bool = attr.ib()
    expiresAt: str = attr.ib()
    resources: List[Resources] = attr.ib()
    audit: Audit = attr.ib()

    @classmethod
    def create(cls, data: dict):
        return cls(
            data.get('id'),
            data.get('name'),
            data.get('email'),
            data.get('status'),
            data.get('inactive'),
            data.get('organizationId'),
            data.get('organizationRoles', []),
            data.get('lastLogin'),
            data.get('region'),
            data.get('timeZone'),
            data.get('enableNotifications'),
            data.get('expiresAt'),
            [
                Resources(
                    r.get('type'),
                    r.get('id'),
                    r.get('roles', [])
                ) for r in data.get('resources', [])
            ],
            Audit(
                data.get("audit", {}).get("createdBy"),
                data.get("audit", {}).get("createdAt"),
                data.get("audit", {}).get("modifiedBy"),
                data.get("audit", {}).get("modifiedAt"),
                data.get("audit", {}).get("version")
            )
        )


@attr.s
class UserOpValue:
    id: Optional[str] = attr.ib(default=None)
    type: Optional[str] = attr.ib(default=None)
    roles: Optional[List[str]] = attr.ib(default=None)


@attr.s
class UserOp:
    op: Optional[str] = attr.ib(default=None)
    path: Optional[str] = attr.ib(default=None)
    value: Optional[UserOpValue] = attr.ib(default=None)


@attr.s
class ProjectOwnership:
    user_op_list: Optional[List[UserOp]] = attr.ib(default=[])

    def add(self, project_id: str):
        opo = UserOp()
        opo.op = "add"
        opo.path = f"/resources/{project_id}"
        opo.value = UserOpValue(id=project_id, type="project", roles=["projectOwner"])
        self.user_op_list.append(opo)

    @property
    def as_dict(self):
        return list(attrs.asdict(o) for o in self.__dict__["user_op_list"])



##
##

import attr
import attrs
from libcapella.logic.common import Audit, not_none


@attr.s
class Project:
    id: str = attr.ib()
    description: str = attr.ib()
    name: str = attr.ib()
    audit: Audit = attr.ib()

    @classmethod
    def create(cls, data: dict):
        return cls(
            data.get("id"),
            data.get("description"),
            data.get("name"),
            Audit(
                data.get("audit", {}).get("createdBy"),
                data.get("audit", {}).get("createdAt"),
                data.get("audit", {}).get("modifiedBy"),
                data.get("audit", {}).get("modifiedAt"),
                data.get("audit", {}).get("version")
            )
        )

    @property
    def as_dict(self):
        # noinspection PyTypeChecker
        return attrs.asdict(self)

    @property
    def as_dict_striped(self):
        result = not_none(self.as_dict)
        if 'audit' in result:
            del result['audit']
        return result


class CapellaProjectBuilder(object):

    def __init__(self,
                 name='default',
                 description='Capella Project'):
        self._name = name
        self._description = description

    def name(self, name: str):
        self._name = name
        return self

    def description(self, description: str):
        self._description = description
        return self

    def build(self) -> Project:
        return Project.create(dict(
            name=self._name,
            description=self._description
        ))



##
##

import logging
import os
import shutil
import configparser
from configparser import SectionProxy
from pathlib import Path
from typing import Union

logger = logging.getLogger('libcapella.config')
logger.addHandler(logging.NullHandler())


class CapellaConfig(object):

    def __init__(self, project: Union[str, None] = None, profile: Union[str, None] = None):
        self.home_dir = Path.home()
        self.config_directory = os.path.join(self.home_dir, '.capella')
        self.config_file = os.path.join(self.config_directory, 'credentials')
        self._profile = profile if profile else 'default'
        self.config_data = configparser.ConfigParser()
        logger.debug(f"using profile: {self._profile}")

        self._api_host = None
        self._token_file = None
        self._token_file_path = os.path.join(self.config_directory, 'default-api-key-token.txt')
        self._organization = None
        self._project = project if project else 'default'
        self._account_email = None
        self._profile_key_id = None
        self._profile_token = None

        logger.debug(f"initializing with credential file: {self.config_file}")
        self.read_config('default')
        if self._profile != 'default':
            self.read_config(self._profile)
        self.read_token()

    def read_config(self, profile: str):
        if not os.path.exists(self.config_file):
            self.write_default_config()

        profile_config = self.read_config_file(profile)
        self._api_host = profile_config.get('api_host', self._api_host)
        self._token_file = profile_config.get('token_file', self._token_file)
        self._token_file_path = os.path.join(self.config_directory, self._token_file)
        self._organization = profile_config.get('organization', self._organization)
        self._project = profile_config.get('project', self._project)
        self._account_email = profile_config.get('account_email', self._account_email)

        logger.debug(f"found token file: {self._token_file}")

    def read_config_file(self, profile: str) -> SectionProxy:
        try:
            self.config_data.read(self.config_file)
            return self.config_data[profile]
        except KeyError:
            raise RuntimeError(f"profile {self._profile} does not exist in config file {self.config_file}")
        except Exception as err:
            raise RuntimeError(f"can not read config file {self.config_file}: {err}")

    def write_default_config(self):
        self.config_data['default'] = {
            'api_host': 'cloudapi.cloud.couchbase.com',
            'token_file': 'default-api-key-token.txt'
        }
        try:
            if not os.path.exists(self.config_directory):
                os.mkdir(self.config_directory)
            with open(self.config_file, 'w') as configfile:
                self.config_data.write(configfile)
        except Exception as err:
            raise RuntimeError(f"can not write config file {self.config_file}: {err}")

    def read_token(self):
        if not os.path.exists(self._token_file_path):
            self.find_token_file()
        self.read_token_file()
        logger.debug(f"using API Key ID: {self._profile_key_id} from {self._token_file}")

    def read_token_file(self):
        if os.path.exists(self._token_file_path):
            try:
                credential_data = dict(line.split(':', 1) for line in open(self._token_file_path))
                self._profile_token = credential_data.get('APIKeyToken').strip()
                self._profile_key_id = credential_data.get('APIKeyId').strip()
            except AttributeError:
                raise RuntimeError(f"token file {self._token_file} does not contain an API key and token")
            except Exception as err:
                raise RuntimeError(f"can not read credential file {self._token_file}: {err}")
        else:
            raise RuntimeError("Please create Capella token file (i.e. $HOME/.capella/default-api-key-token.txt)")

    def find_token_file(self):
        download_dir = os.path.join(Path.home(), 'Downloads')

        if os.path.exists(os.path.join(download_dir, self.token_file)):
            shutil.copy(os.path.join(download_dir, self.token_file), self._token_file_path)
        elif os.path.exists(os.path.join(self.home_dir, self.token_file)):
            shutil.copy(os.path.join(self.home_dir, self.token_file), self._token_file_path)

    @property
    def api_host(self):
        return self._api_host

    @property
    def token_file(self):
        return self._token_file

    @property
    def organization(self):
        return self._organization

    @property
    def project(self):
        return self._project

    @property
    def account_email(self):
        return self._account_email

    @property
    def token(self):
        return self._profile_token

    @property
    def key_id(self):
        return self._profile_key_id

    @property
    def profile(self):
        return self._profile
